@@ -11,7 +11,7 @@
                             Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>,
                             Leonhard Wimmer <leo@mediatomb.cc>
 
-    Copyright (C) 2016-2024 Gerbera Contributors
+    Copyright (C) 2016-2025 Gerbera Contributors
 
     MediaTomb is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2
@@ -32,10 +32,21 @@
 /// \file taglib_handler.cc
 
 #ifdef HAVE_TAGLIB
-#define LOG_FAC log_facility_t::taglib
+#define GRB_LOG_FAC GrbLogFacility::taglib
 
 #include "taglib_handler.h" // API
 
+#include "cds/cds_item.h"
+#include "config/config.h"
+#include "config/config_val.h"
+#include "exceptions.h"
+#include "iohandler/mem_io_handler.h"
+#include "metadata_enums.h"
+#include "util/grb_time.h"
+#include "util/mime.h"
+#include "util/string_converter.h"
+#include "util/tools.h"
+
 #include <aifffile.h>
 #include <apefile.h>
 #include <asffile.h>
@@ -47,6 +58,7 @@
 #include <oggflacfile.h>
 #include <opusfile.h>
 #include <speexfile.h>
+#include <tdebuglistener.h>
 #include <textidentificationframe.h>
 #include <tfilestream.h>
 #include <tiostream.h>
@@ -54,32 +66,64 @@
 #include <vorbisfile.h>
 #include <wavpackfile.h>
 
-#include "cds/cds_item.h"
-#include "config/config_manager.h"
-#include "iohandler/mem_io_handler.h"
-#include "metadata_enums.h"
-#include "util/grb_time.h"
-#include "util/mime.h"
-#include "util/string_converter.h"
-#include "util/tools.h"
+#if TAGLIB_MAJOR_VERSION >= 2
+#include <dsffile.h>
+#include <mp4itemfactory.h>
+#include <wavfile.h>
+#endif
+
+class GerberaTagLibDebugListener : public TagLib::DebugListener {
+private:
+    GerberaTagLibDebugListener()
+    {
+        // Install listener.
+        TagLib::setDebugListener(this);
+    }
+
+    void printMessage(const TagLib::String& msg) override
+    {
+        // Remove trailing newlines.
+        log_debug("{}", msg.stripWhiteSpace().to8Bit(true));
+    }
+    static GerberaTagLibDebugListener grbListener;
+};
+
+GerberaTagLibDebugListener GerberaTagLibDebugListener::grbListener;
 
 TagLibHandler::TagLibHandler(const std::shared_ptr<Context>& context)
-    : MetadataHandler(context)
+    : MediaMetadataHandler(context,
+          ConfigVal::IMPORT_LIBOPTS_ID3_ENABLED,
+          ConfigVal::IMPORT_LIBOPTS_ID3_METADATA_TAGS_LIST,
+          ConfigVal::IMPORT_LIBOPTS_ID3_AUXDATA_TAGS_LIST,
+          ConfigVal::IMPORT_LIBOPTS_ID3_COMMENT_ENABLED,
+          ConfigVal::IMPORT_LIBOPTS_ID3_COMMENT_LIST)
 {
-    entrySeparator = this->config->getOption(CFG_IMPORT_LIBOPTS_ENTRY_SEP);
-    legacyEntrySeparator = this->config->getOption(CFG_IMPORT_LIBOPTS_ENTRY_LEGACY_SEP);
+    entrySeparator = this->config->getOption(ConfigVal::IMPORT_LIBOPTS_ENTRY_SEP);
+    legacyEntrySeparator = this->config->getOption(ConfigVal::IMPORT_LIBOPTS_ENTRY_LEGACY_SEP);
     trimStringInPlace(legacyEntrySeparator);
     if (legacyEntrySeparator.empty())
         legacyEntrySeparator = "\n";
-    specialPropertyMap = this->config->getDictionaryOption(CFG_IMPORT_LIBOPTS_ID3_METADATA_TAGS_LIST);
 }
 
-void TagLibHandler::addField(MetadataFields field, const TagLib::File& file, const TagLib::Tag* tag, const std::shared_ptr<CdsItem>& item) const
+bool TagLibHandler::isSupported(
+    const std::string& contentType,
+    bool isOggTheora,
+    const std::string& mimeType,
+    ObjectType mediaType)
+{
+    return contentType == CONTENT_TYPE_MP3 || (contentType == CONTENT_TYPE_OGG && !isOggTheora) || contentType == CONTENT_TYPE_WMA || contentType == CONTENT_TYPE_WAVPACK || contentType == CONTENT_TYPE_FLAC || contentType == CONTENT_TYPE_PCM || contentType == CONTENT_TYPE_AIFF || contentType == CONTENT_TYPE_APE || contentType == CONTENT_TYPE_MP4 || contentType == CONTENT_TYPE_DSD;
+}
+
+void TagLibHandler::addField(
+    MetadataFields field,
+    const TagLib::File& file,
+    const TagLib::Tag* tag,
+    const std::shared_ptr<CdsItem>& item) const
 {
     if (!tag || tag->isEmpty())
         return;
 
-    auto sc = StringConverter::i2i(config); // sure is sure
+    auto sc = converterManager->i2i(); // sure is sure
 
     std::vector<std::string> value;
 
@@ -151,21 +195,29 @@ void TagLibHandler::addField(MetadataFields field, const TagLib::File& file, con
     if (!value.empty()) {
         for (auto&& entry : value) {
             trimStringInPlace(entry);
-            if (!entry.empty())
-                item->addMetaData(field, sc->convert(entry));
+            if (!entry.empty()) {
+                auto [val, err] = sc->convert(entry);
+                if (!err.empty()) {
+                    log_warning("{}: {}", item->getLocation().string(), err);
+                }
+                item->addMetaData(field, val);
+                log_debug("Setting metadata on item: {} = {}", field, val);
+            }
         }
-        // [log_debug("Setting metadata on item: {} = {}", field, sc->convert(value).c_str());]
     }
 }
 
-void TagLibHandler::addSpecialFields(const TagLib::File& file, const TagLib::Tag* tag, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::addSpecialFields(
+    const TagLib::File& file,
+    const TagLib::Tag* tag,
+    const std::shared_ptr<CdsItem>& item) const
 {
     if (!tag || tag->isEmpty())
         return;
 
-    auto sc = StringConverter::i2i(config); // sure is sure
+    auto sc = converterManager->i2i(); // sure is sure
 
-    for (auto&& [key, meta] : specialPropertyMap) {
+    for (auto&& [key, meta] : metaTags) {
         auto list = file.properties()[key];
         if (list.isEmpty())
             return;
@@ -173,14 +225,24 @@ void TagLibHandler::addSpecialFields(const TagLib::File& file, const TagLib::Tag
             for (auto&& entrySeg : val.split(legacyEntrySeparator)) {
                 std::string entry = entrySeg.to8Bit(true);
                 trimStringInPlace(entry);
-                if (!entry.empty())
-                    item->addMetaData(meta, sc->convert(entry));
+                if (!entry.empty()) {
+                    auto [val, err] = sc->convert(entry);
+                    if (!err.empty()) {
+                        log_warning("{}: {}", item->getLocation().string(), err);
+                    }
+                    item->addMetaData(meta, val);
+                }
             }
         }
     }
 }
 
-void TagLibHandler::populateGenericTags(const std::shared_ptr<CdsItem>& item, const TagLib::File& file) const
+void TagLibHandler::populateGenericTags(
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res,
+    const TagLib::File& file,
+    const TagLib::PropertyMap& propertyMap,
+    const std::shared_ptr<StringConverter>& sc) const
 {
     if (!file.tag())
         return;
@@ -195,34 +257,46 @@ void TagLibHandler::populateGenericTags(const std::shared_ptr<CdsItem>& item, co
     if (!audioProps)
         return;
 
-    auto res = item->getResource(ContentHandler::DEFAULT);
-
-    // UPnP bitrate is in bytes/second
-    int temp = audioProps->bitrate() * 1024 / 8; // kbit/second -> byte/second
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::BITRATE, fmt::to_string(temp));
+    {
+        // UPnP bitrate is in bytes/second
+        int temp = audioProps->bitrate() * 1024 / 8; // kbit/second -> byte/second
+        if (temp > 0) {
+            res->addAttribute(ResourceAttribute::BITRATE, fmt::to_string(temp));
+        }
     }
 
-    temp = audioProps->lengthInMilliseconds();
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::DURATION, millisecondsToHMSF(temp));
+    {
+        int temp = audioProps->lengthInMilliseconds();
+        if (temp > 0) {
+            res->addAttribute(ResourceAttribute::DURATION, millisecondsToHMSF(temp));
+        }
     }
 
-    temp = audioProps->sampleRate();
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::SAMPLEFREQUENCY, fmt::to_string(temp));
+    {
+        int temp = audioProps->sampleRate();
+        if (temp > 0) {
+            res->addAttribute(ResourceAttribute::SAMPLEFREQUENCY, fmt::to_string(temp));
+        }
     }
 
-    temp = audioProps->channels();
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::NRAUDIOCHANNELS, fmt::to_string(temp));
+    {
+        int temp = audioProps->channels();
+        if (temp > 0) {
+            res->addAttribute(ResourceAttribute::NRAUDIOCHANNELS, fmt::to_string(temp));
+        }
     }
+
+    populateAuxTags(item, propertyMap, sc);
+    if (item->getMetaData(MetadataFields::M_DESCRIPTION).empty() && isCommentEnabled)
+        makeComment(item, propertyMap, sc);
 }
 
-void TagLibHandler::populateAuxTags(const std::shared_ptr<CdsItem>& item, const TagLib::PropertyMap& propertyMap, const std::unique_ptr<StringConverter>& sc) const
+void TagLibHandler::populateAuxTags(
+    const std::shared_ptr<CdsItem>& item,
+    const TagLib::PropertyMap& propertyMap,
+    const std::shared_ptr<StringConverter>& sc) const
 {
-    auto auxTagsList = config->getArrayOption(CFG_IMPORT_LIBOPTS_ID3_AUXDATA_TAGS_LIST);
-    for (auto&& desiredTag : auxTagsList) {
+    for (auto&& desiredTag : auxTags) {
         if (desiredTag.empty()) {
             continue;
         }
@@ -232,57 +306,93 @@ void TagLibHandler::populateAuxTags(const std::shared_ptr<CdsItem>& item, const
             if (property.isEmpty())
                 continue;
 
-            auto val = property.toString(entrySeparator);
+            auto entry = property.toString(entrySeparator);
             if (!legacyEntrySeparator.empty())
-                val = val.split(legacyEntrySeparator).toString(entrySeparator);
-            std::string value(val.to8Bit(true));
-            value = sc->convert(value);
-            log_debug("Adding auxdata: {} with value {}", desiredTag.c_str(), value.c_str());
-            item->setAuxData(desiredTag, value);
+                entry = entry.split(legacyEntrySeparator).toString(entrySeparator);
+            std::string value(entry.to8Bit(true));
+            auto [val, err] = sc->convert(value);
+            if (!err.empty()) {
+                log_warning("{}: {}", item->getLocation().string(), err);
+            }
+            log_debug("Adding auxdata: {} with value {}", desiredTag, val);
+            item->setAuxData(desiredTag, val);
         }
     }
 }
 
-/// \brief read metadata from file and add to object
-/// \param obj Object to handle
-void TagLibHandler::fillMetadata(const std::shared_ptr<CdsObject>& obj)
+void TagLibHandler::makeComment(
+    const std::shared_ptr<CdsItem>& item,
+    const TagLib::PropertyMap& propertyMap,
+    const std::shared_ptr<StringConverter>& sc) const
 {
-    auto item = std::dynamic_pointer_cast<CdsItem>(obj);
-    if (!item)
-        return;
+    std::vector<std::string> snippets;
+    for (auto&& [label, desiredTag] : commentMap) {
+        if (desiredTag.empty()) {
+            continue;
+        }
 
-    auto mappings = config->getDictionaryOption(CFG_IMPORT_MAPPINGS_MIMETYPE_TO_CONTENTTYPE_LIST);
-    std::string contentType = getValueOrDefault(mappings, item->getMimeType());
+        if (propertyMap.contains(desiredTag.c_str())) {
+            auto&& property = propertyMap[desiredTag.c_str()];
+            if (property.isEmpty())
+                continue;
 
-    auto fs = TagLib::FileStream(item->getLocation().c_str(), true); // true = Read only
-
-    if (contentType == CONTENT_TYPE_MP3) {
-        extractMP3(fs, item);
-    } else if (contentType == CONTENT_TYPE_FLAC) {
-        extractFLAC(fs, item);
-    } else if (contentType == CONTENT_TYPE_MP4) {
-        extractMP4(fs, item);
-    } else if (contentType == CONTENT_TYPE_OGG) {
-        extractOgg(fs, item);
-    } else if (contentType == CONTENT_TYPE_APE) {
-        extractAPE(fs, item);
-    } else if (contentType == CONTENT_TYPE_WMA) {
-        extractASF(fs, item);
-    } else if (contentType == CONTENT_TYPE_WAVPACK) {
-        extractWavPack(fs, item);
-    } else if (contentType == CONTENT_TYPE_AIFF) {
-        extractAiff(fs, item);
-    } else {
-        log_warning("TagLibHandler {}: Does not handle the {} content type", item->getLocation().c_str(), contentType.c_str());
+            auto entry = property.toString(entrySeparator);
+            if (!legacyEntrySeparator.empty())
+                entry = entry.split(legacyEntrySeparator).toString(entrySeparator);
+            std::string value(entry.to8Bit(true));
+            auto [val, err] = sc->convert(value);
+            if (!err.empty()) {
+                log_warning("{}: {}", item->getLocation().string(), err);
+            }
+            log_debug("Adding auxdata: {} with value {}", desiredTag, val);
+            snippets.push_back(fmt::format("{}: {}", label, val));
+        }
     }
-    log_debug("TagLib handler done.");
+    item->addMetaData(MetadataFields::M_DESCRIPTION, fmt::format("{}", fmt::join(snippets, ", ")));
+    log_debug("Fabricated Comment: {}", fmt::format("{}", fmt::join(snippets, ", ")));
 }
 
-bool TagLibHandler::isValidArtworkContentType(std::string_view artMimetype)
+/// \brief read metadata from file and add to object
+/// \param obj Object to handle
+bool TagLibHandler::fillMetadata(const std::shared_ptr<CdsObject>& obj)
 {
-    // saw that simply "PNG" was used with some mp3's, so mimetype setting
-    // was probably invalid
-    return artMimetype.find('/') != std::string_view::npos;
+    auto item = std::dynamic_pointer_cast<CdsItem>(obj);
+    if (!item || !isEnabled)
+        return false;
+
+    auto res = item->getResource(ContentHandler::DEFAULT);
+    std::string contentType = getValueOrDefault(mimeContentTypeMappings, item->getMimeType());
+
+    log_debug("Reading {}, content type {}", item->getLocation().c_str(), contentType);
+    auto fs = TagLib::FileStream(item->getLocation().c_str(), true); // true = Read only
+
+    bool result = true;
+    if (contentType == CONTENT_TYPE_MP3) {
+        extractMP3(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_FLAC) {
+        extractFLAC(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_MP4) {
+        extractMP4(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_OGG) {
+        extractOgg(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_APE) {
+        extractAPE(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_WMA) {
+        extractASF(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_WAVPACK) {
+        extractWavPack(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_DSD) {
+        extractDSF(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_AIFF) {
+        extractAiff(fs, item, res);
+    } else if (contentType == CONTENT_TYPE_PCM) {
+        extractPcm(fs, item, res);
+    } else {
+        log_warning("TagLibHandler: File '{}' has unhandled content type '{}'", item->getLocation().c_str(), contentType.c_str());
+        result = false;
+    }
+    log_debug("Done {}", item->getLocation().c_str());
+    return result;
 }
 
 std::string TagLibHandler::getContentTypeFromByteVector(const TagLib::ByteVector& data) const
@@ -295,31 +405,19 @@ std::string TagLibHandler::getContentTypeFromByteVector(const TagLib::ByteVector
 #endif
 }
 
-void TagLibHandler::addArtworkResource(const std::shared_ptr<CdsItem>& item, const std::string& artMimetype)
-{
-    // if we could not determine the mimetype, then there is no
-    // point to add the resource - it's probably garbage
-    log_debug("Found artwork of type {} in file {}", artMimetype.c_str(), item->getLocation().c_str());
-
-    if (artMimetype != MIMETYPE_DEFAULT) {
-        auto resource = std::make_shared<CdsResource>(ContentHandler::ID3, ResourcePurpose::Thumbnail);
-        resource->addAttribute(ResourceAttribute::PROTOCOLINFO, renderProtocolInfo(artMimetype));
-        item->addResource(resource);
-    }
-}
-
 /// \brief stream content of object or resource to client
 /// \param obj Object to stream
 /// \param resource the resource
 /// \return iohandler to stream to client
-std::unique_ptr<IOHandler> TagLibHandler::serveContent(const std::shared_ptr<CdsObject>& obj, const std::shared_ptr<CdsResource>& resource)
+std::unique_ptr<IOHandler> TagLibHandler::serveContent(
+    const std::shared_ptr<CdsObject>& obj,
+    const std::shared_ptr<CdsResource>& resource)
 {
     auto item = std::dynamic_pointer_cast<CdsItem>(obj);
     if (!item) // not streamable
         return nullptr;
 
-    auto mappings = config->getDictionaryOption(CFG_IMPORT_MAPPINGS_MIMETYPE_TO_CONTENTTYPE_LIST);
-    std::string contentType = getValueOrDefault(mappings, item->getMimeType());
+    std::string contentType = getValueOrDefault(mimeContentTypeMappings, item->getMimeType());
     auto itemLocation = item->getLocation().c_str();
 
     auto roStream = TagLib::FileStream(itemLocation, true); // Open read only
@@ -327,7 +425,7 @@ std::unique_ptr<IOHandler> TagLibHandler::serveContent(const std::shared_ptr<Cds
     if (contentType == CONTENT_TYPE_MP3) {
         // stream album art from MP3 file
 #if TAGLIB_MAJOR_VERSION >= 2
-        auto f = TagLib::MPEG::File(&roStream);
+        auto f = TagLib::MPEG::File(&roStream, true, TagLib::MPEG::Properties::Average, TagLib::ID3v2::FrameFactory::instance());
 #else
         auto f = TagLib::MPEG::File(&roStream, TagLib::ID3v2::FrameFactory::instance());
 #endif
@@ -349,10 +447,35 @@ std::unique_ptr<IOHandler> TagLibHandler::serveContent(const std::shared_ptr<Cds
 
         return std::make_unique<MemIOHandler>(art->picture().data(), art->picture().size());
     }
+    if (contentType == CONTENT_TYPE_DSD) {
+#if TAGLIB_MAJOR_VERSION >= 2
+        auto f = TagLib::DSF::File(&roStream);
+
+        if (!f.isValid())
+            throw_std_runtime_error("Could not open file {} as DSF", itemLocation);
+
+        if (!f.tag())
+            throw_std_runtime_error("DSF resource {} has no ID3 information", itemLocation);
+
+        auto&& list = f.tag()->frameList("APIC");
+        if (list.isEmpty())
+            throw_std_runtime_error("DSF resource {} has no album information", itemLocation);
+
+        auto art = dynamic_cast<TagLib::ID3v2::AttachedPictureFrame*>(list.front());
+        if (!art) {
+            return nullptr;
+        }
+
+        return std::make_unique<MemIOHandler>(art->picture().data(), art->picture().size());
+#else
+        log_warning("For DSF support in file '{}' TaglibHandler needs to be built with taglib 2 and above", itemLocation);
+        return nullptr;
+#endif
+    }
     if (contentType == CONTENT_TYPE_FLAC) {
         // stream album art from FLAC file
 #if TAGLIB_MAJOR_VERSION >= 2
-        auto f = TagLib::FLAC::File(&roStream);
+        auto f = TagLib::FLAC::File(&roStream, true, TagLib::MPEG::Properties::Average, TagLib::ID3v2::FrameFactory::instance());
 #else
         auto f = TagLib::FLAC::File(&roStream, TagLib::ID3v2::FrameFactory::instance());
 #endif
@@ -440,10 +563,13 @@ std::unique_ptr<IOHandler> TagLibHandler::serveContent(const std::shared_ptr<Cds
     throw_std_runtime_error("File {} has unsupported content type {}", itemLocation, contentType);
 }
 
-void TagLibHandler::extractMP3(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractMP3(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
 #if TAGLIB_MAJOR_VERSION >= 2
-    auto mp3 = TagLib::MPEG::File(&roStream);
+    auto mp3 = TagLib::MPEG::File(&roStream, true, TagLib::MPEG::Properties::Average, TagLib::ID3v2::FrameFactory::instance());
 #else
     auto mp3 = TagLib::MPEG::File(&roStream, TagLib::ID3v2::FrameFactory::instance());
 #endif
@@ -451,21 +577,20 @@ void TagLibHandler::extractMP3(TagLib::IOStream& roStream, const std::shared_ptr
         log_info("TagLibHandler {}: does not appear to be a valid mp3 file", item->getLocation().c_str());
         return;
     }
-    populateGenericTags(item, mp3);
+
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, mp3, mp3.hasID3v2Tag() ? mp3.ID3v2Tag()->properties() : mp3.properties(), sc);
 
     if (!mp3.hasID3v2Tag()) {
         log_debug("{}: has no IDv2 tags", item->getLocation().c_str());
         return;
     }
 
-    auto sc = StringConverter::i2i(config);
-
     auto&& frameListMap = mp3.ID3v2Tag()->frameListMap();
     // http://id3.org/id3v2.4.0-frames "4.2.6. User defined text information frame"
     bool hasTXXXFrames = frameListMap.contains("TXXX");
 
-    std::vector<std::string> auxTagsList = config->getArrayOption(CFG_IMPORT_LIBOPTS_ID3_AUXDATA_TAGS_LIST);
-    for (auto&& desiredFrame : auxTagsList) {
+    for (auto&& desiredFrame : auxTags) {
         if (desiredFrame.empty()) {
             continue;
         }
@@ -481,11 +606,21 @@ void TagLibHandler::extractMP3(TagLib::IOStream& roStream, const std::shared_ptr
                 if (!textFrame)
                     continue;
                 for (auto&& field : textFrame->fieldList()) {
-                    if (legacyEntrySeparator.empty())
-                        content.push_back(sc->convert(field.to8Bit(true)));
-                    else
-                        for (auto&& val : field.split(legacyEntrySeparator))
-                            content.push_back(sc->convert(val.to8Bit(true)));
+                    if (legacyEntrySeparator.empty()) {
+                        auto [val, err] = sc->convert(field.to8Bit(true));
+                        if (!err.empty()) {
+                            log_warning("{}: {}", item->getLocation().string(), err);
+                        }
+                        content.push_back(val);
+                    } else {
+                        for (auto&& fval : field.split(legacyEntrySeparator)) {
+                            auto [val, err] = sc->convert(fval.to8Bit(true));
+                            if (!err.empty()) {
+                                log_warning("{}: {}", item->getLocation().string(), err);
+                            }
+                            content.push_back(val);
+                        }
+                    }
                 }
             }
             if (!content.empty()) {
@@ -494,7 +629,7 @@ void TagLibHandler::extractMP3(TagLib::IOStream& roStream, const std::shared_ptr
             }
         } else if (hasTXXXFrames && startswith(desiredFrame, "TXXX:")) {
             auto&& frameList = frameListMap["TXXX"];
-            // log_debug("TXXX Frame list has {} elements", frameList.size());
+            log_debug("TXXX Frame list has {} elements", frameList.size());
 
             std::string desiredSubTag = desiredFrame.substr(5);
             if (desiredSubTag.empty())
@@ -509,12 +644,25 @@ void TagLibHandler::extractMP3(TagLib::IOStream& roStream, const std::shared_ptr
                 for (auto&& field : textFrame->fieldList()) {
                     if (subTag.empty()) {
                         // first element is subTag name
-                        subTag = sc->convert(field.to8Bit(true));
+                        auto [val, err] = sc->convert(field.to8Bit(true));
+                        if (!err.empty()) {
+                            log_warning("{}: {}", item->getLocation().string(), err);
+                        }
+                        subTag = val;
                     } else if (legacyEntrySeparator.empty()) {
-                        content.push_back(sc->convert(field.to8Bit(true)));
+                        auto [val, err] = sc->convert(field.to8Bit(true));
+                        if (!err.empty()) {
+                            log_warning("{}: {}", item->getLocation().string(), err);
+                        }
+                        content.push_back(val);
                     } else {
-                        for (auto&& val : field.split(legacyEntrySeparator))
-                            content.push_back(sc->convert(val.to8Bit(true)));
+                        for (auto&& fval : field.split(legacyEntrySeparator)) {
+                            auto [val, err] = sc->convert(fval.to8Bit(true));
+                            if (!err.empty()) {
+                                log_warning("{}: {}", item->getLocation().string(), err);
+                            }
+                            content.push_back(val);
+                        }
                     }
                 }
                 log_debug("TXXX Tag: {}", subTag);
@@ -539,12 +687,17 @@ void TagLibHandler::extractMP3(TagLib::IOStream& roStream, const std::shared_ptr
         }
 
         auto pic = art->picture();
-        std::string artMimetype = sc->convert(art->mimeType().to8Bit(true));
+        auto [artMimetype, err] = sc->convert(art->mimeType().to8Bit(true));
+        if (!err.empty()) {
+            log_warning("{}: {}", item->getLocation().string(), err);
+        }
         if (!isValidArtworkContentType(artMimetype)) {
             artMimetype = getContentTypeFromByteVector(pic);
         }
 
-        addArtworkResource(item, artMimetype);
+        auto resource = addArtworkResource(item, ContentHandler::ID3, artMimetype);
+        if (resource)
+            resource->addAttribute(ResourceAttribute::SIZE, fmt::to_string(pic.size()));
     }
 }
 
@@ -570,7 +723,10 @@ std::unique_ptr<TagLib::File> TagLibHandler::getOggFile(TagLib::IOStream& ioStre
     return nullptr;
 }
 
-void TagLibHandler::extractOgg(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractOgg(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
     auto oggFile = getOggFile(roStream);
 
@@ -578,14 +734,13 @@ void TagLibHandler::extractOgg(TagLib::IOStream& roStream, const std::shared_ptr
         log_info("TagLibHandler {}: does not appear to be a valid ogg file", item->getLocation().c_str());
         return;
     }
-    populateGenericTags(item, *oggFile);
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, *oggFile, oggFile->properties(), sc);
 
-    if (!oggFile->tag())
+    if (!oggFile->tag()) {
+        log_debug("{}: has no tags", item->getLocation().c_str());
         return;
-
-    auto sc = StringConverter::i2i(config);
-    auto propertyMap = oggFile->properties();
-    populateAuxTags(item, propertyMap, sc);
+    }
 
     // Vorbis uses the FLAC binary picture structure...
     // https://wiki.xiph.org/VorbisComment#Cover_art
@@ -597,14 +752,20 @@ void TagLibHandler::extractOgg(TagLib::IOStream& roStream, const std::shared_ptr
     const TagLib::FLAC::Picture* pic = picList.front();
     const TagLib::ByteVector& data = pic->data();
 
-    std::string artMimetype = sc->convert(pic->mimeType().to8Bit(true));
+    auto [artMimetype, err] = sc->convert(pic->mimeType().to8Bit(true));
+    if (!err.empty()) {
+        log_warning("{}: {}", item->getLocation().string(), err);
+    }
     if (!isValidArtworkContentType(artMimetype)) {
         artMimetype = getContentTypeFromByteVector(data);
     }
-    addArtworkResource(item, artMimetype);
+    addArtworkResource(item, ContentHandler::ID3, artMimetype);
 }
 
-void TagLibHandler::extractASF(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractASF(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
     auto asf = TagLib::ASF::File(&roStream);
 
@@ -612,18 +773,10 @@ void TagLibHandler::extractASF(TagLib::IOStream& roStream, const std::shared_ptr
         log_info("TagLibHandler {}: does not appear to be a valid asf/wma file", item->getLocation().c_str());
         return;
     }
-    populateGenericTags(item, asf);
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, asf, asf.properties(), sc);
 
-    auto sc = StringConverter::i2i(config);
-    auto propertyMap = asf.properties();
-    populateAuxTags(item, propertyMap, sc);
-
-    auto audioProps = asf.audioProperties();
-    auto temp = audioProps->bitsPerSample();
-    auto res = item->getResource(ContentHandler::DEFAULT);
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::BITS_PER_SAMPLE, fmt::to_string(temp));
-    }
+    setBitsPerSample(item, res, asf);
 
     const TagLib::ASF::AttributeListMap& attrListMap = asf.tag()->attributeListMap();
     if (attrListMap.contains("WM/Picture")) {
@@ -635,18 +788,24 @@ void TagLibHandler::extractASF(TagLib::IOStream& roStream, const std::shared_ptr
         if (!wmpic.isValid())
             return;
 
-        std::string artMimetype = sc->convert(wmpic.mimeType().to8Bit(true));
+        auto [artMimetype, err] = sc->convert(wmpic.mimeType().to8Bit(true));
+        if (!err.empty()) {
+            log_warning("{}: {}", item->getLocation().string(), err);
+        }
         if (!isValidArtworkContentType(artMimetype)) {
             artMimetype = getContentTypeFromByteVector(wmpic.picture());
         }
-        addArtworkResource(item, artMimetype);
+        addArtworkResource(item, ContentHandler::ID3, artMimetype);
     }
 }
 
-void TagLibHandler::extractFLAC(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractFLAC(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
 #if TAGLIB_MAJOR_VERSION >= 2
-    auto flac = TagLib::FLAC::File(&roStream);
+    auto flac = TagLib::FLAC::File(&roStream, true, TagLib::MPEG::Properties::Average, TagLib::ID3v2::FrameFactory::instance());
 #else
     auto flac = TagLib::FLAC::File(&roStream, TagLib::ID3v2::FrameFactory::instance());
 #endif
@@ -655,18 +814,10 @@ void TagLibHandler::extractFLAC(TagLib::IOStream& roStream, const std::shared_pt
         log_info("TagLibHandler {}: does not appear to be a valid flac file", item->getLocation().c_str());
         return;
     }
-    populateGenericTags(item, flac);
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, flac, flac.properties(), sc);
 
-    auto sc = StringConverter::i2i(config);
-    auto propertyMap = flac.properties();
-    populateAuxTags(item, propertyMap, sc);
-
-    auto audioProps = flac.audioProperties();
-    auto temp = audioProps->bitsPerSample();
-    auto res = item->getResource(ContentHandler::DEFAULT);
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::BITS_PER_SAMPLE, fmt::to_string(temp));
-    }
+    setBitsPerSample(item, res, flac);
 
     if (flac.pictureList().isEmpty()) {
         log_debug("TagLibHandler: flac resource has no picture information");
@@ -675,14 +826,20 @@ void TagLibHandler::extractFLAC(TagLib::IOStream& roStream, const std::shared_pt
     const TagLib::FLAC::Picture* pic = flac.pictureList().front();
     const TagLib::ByteVector& data = pic->data();
 
-    std::string artMimetype = sc->convert(pic->mimeType().to8Bit(true));
+    auto [artMimetype, err] = sc->convert(pic->mimeType().to8Bit(true));
+    if (!err.empty()) {
+        log_warning("{}: {}", item->getLocation().string(), err);
+    }
     if (!isValidArtworkContentType(artMimetype)) {
         artMimetype = getContentTypeFromByteVector(data);
     }
-    addArtworkResource(item, artMimetype);
+    addArtworkResource(item, ContentHandler::ID3, artMimetype);
 }
 
-void TagLibHandler::extractAPE(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractAPE(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
     auto ape = TagLib::APE::File(&roStream);
 
@@ -690,21 +847,16 @@ void TagLibHandler::extractAPE(TagLib::IOStream& roStream, const std::shared_ptr
         log_info("TagLibHandler {}: does not appear to be a valid APE file", item->getLocation().c_str());
         return;
     }
-    populateGenericTags(item, ape);
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, ape, ape.properties(), sc);
 
-    auto sc = StringConverter::i2i(config);
-    auto propertyMap = ape.properties();
-    populateAuxTags(item, propertyMap, sc);
-
-    auto audioProps = ape.audioProperties();
-    auto temp = audioProps->bitsPerSample();
-    auto res = item->getResource(ContentHandler::DEFAULT);
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::BITS_PER_SAMPLE, fmt::to_string(temp));
-    }
+    setBitsPerSample(item, res, ape);
 }
 
-void TagLibHandler::extractWavPack(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractWavPack(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
     auto wavpack = TagLib::WavPack::File(&roStream);
 
@@ -712,46 +864,79 @@ void TagLibHandler::extractWavPack(TagLib::IOStream& roStream, const std::shared
         log_info("TagLibHandler {}: does not appear to be a valid WavPack file", item->getLocation().c_str());
         return;
     }
-    populateGenericTags(item, wavpack);
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, wavpack, wavpack.properties(), sc);
 
-    auto sc = StringConverter::i2i(config);
-    auto propertyMap = wavpack.properties();
-    populateAuxTags(item, propertyMap, sc);
-
-    auto audioProps = wavpack.audioProperties();
-    auto temp = audioProps->bitsPerSample();
-    auto res = item->getResource(ContentHandler::DEFAULT);
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::BITS_PER_SAMPLE, fmt::to_string(temp));
-    }
+    setBitsPerSample(item, res, wavpack);
 }
 
-void TagLibHandler::extractMP4(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractDSF(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
+#if TAGLIB_MAJOR_VERSION >= 2
+    auto dsf = TagLib::DSF::File(&roStream);
+
+    if (!dsf.isValid()) {
+        log_info("TagLibHandler {}: does not appear to be a valid dsf file", item->getLocation().c_str());
+        return;
+    }
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, dsf, dsf.properties(), sc);
+
+    setBitsPerSample(item, res, dsf);
+
+    auto&& apicFrameList = dsf.tag()->frameList("APIC");
+    if (!apicFrameList.isEmpty()) {
+        auto art = dynamic_cast<const TagLib::ID3v2::AttachedPictureFrame*>(apicFrameList.front());
+        if (!art) {
+            return;
+        }
+
+        auto pic = art->picture();
+        auto [artMimetype, err] = sc->convert(art->mimeType().to8Bit(true));
+        if (!err.empty()) {
+            log_warning("{}: {}", item->getLocation().string(), err);
+        }
+        if (!isValidArtworkContentType(artMimetype)) {
+            artMimetype = getContentTypeFromByteVector(pic);
+        }
+
+        auto resource = addArtworkResource(item, ContentHandler::ID3, artMimetype);
+        if (resource)
+            resource->addAttribute(ResourceAttribute::SIZE, fmt::to_string(pic.size()));
+    }
+#else
+    log_warning("For DSF support in file '{}' TaglibHandler needs to be built with taglib 2 and above", item->getLocation().c_str());
+#endif
+}
+
+void TagLibHandler::extractMP4(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
+{
+#if TAGLIB_MAJOR_VERSION >= 2
+    auto mp4 = TagLib::MP4::File(&roStream, true, TagLib::MP4::Properties::Average, TagLib::MP4::ItemFactory::instance());
+#else
     auto mp4 = TagLib::MP4::File(&roStream);
+#endif
 
     if (!mp4.isValid()) {
         log_info("TagLibHandler {}: does not appear to be a valid mp4 file", item->getLocation().c_str());
         return;
     }
 
-    populateGenericTags(item, mp4);
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, mp4, mp4.hasMP4Tag() ? mp4.tag()->properties() : mp4.properties(), sc);
 
     if (!mp4.hasMP4Tag()) {
         log_debug("TagLibHandler {}: mp4 file has no tag information", item->getLocation().c_str());
         return;
     }
 
-    auto sc = StringConverter::i2i(config);
-    auto propertyMap = mp4.tag()->properties();
-    populateAuxTags(item, propertyMap, sc);
-
-    auto audioProps = mp4.audioProperties();
-    auto temp = audioProps->bitsPerSample();
-    auto res = item->getResource(ContentHandler::DEFAULT);
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::BITS_PER_SAMPLE, fmt::to_string(temp));
-    }
+    setBitsPerSample(item, res, mp4);
 
     if (mp4.tag()->contains("covr")) {
         auto coverItem = mp4.tag()->item("covr");
@@ -765,7 +950,7 @@ void TagLibHandler::extractMP4(TagLib::IOStream& roStream, const std::shared_ptr
         const auto& coverArt = coverArtList.front();
         auto artMimetype = getContentTypeFromByteVector(coverArt.data());
         if (!artMimetype.empty()) {
-            addArtworkResource(item, artMimetype);
+            addArtworkResource(item, ContentHandler::ID3, artMimetype);
         }
     } else {
         log_debug("TagLibHandler {}: mp4 file has no 'covr' item",
@@ -773,7 +958,10 @@ void TagLibHandler::extractMP4(TagLib::IOStream& roStream, const std::shared_ptr
     }
 }
 
-void TagLibHandler::extractAiff(TagLib::IOStream& roStream, const std::shared_ptr<CdsItem>& item) const
+void TagLibHandler::extractAiff(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
 {
     auto aiff = TagLib::RIFF::AIFF::File(&roStream);
 
@@ -781,17 +969,44 @@ void TagLibHandler::extractAiff(TagLib::IOStream& roStream, const std::shared_pt
         log_info("TagLibHandler {}: does not appear to be a valid AIFF file", item->getLocation().c_str());
         return;
     }
-    populateGenericTags(item, aiff);
 
-    auto sc = StringConverter::i2i(config);
-    auto propertyMap = aiff.properties();
-    populateAuxTags(item, propertyMap, sc);
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, aiff, aiff.properties(), sc);
 
-    auto audioProps = aiff.audioProperties();
-    auto temp = audioProps->bitsPerSample();
-    auto res = item->getResource(ContentHandler::DEFAULT);
-    if (temp > 0) {
-        res->addAttribute(ResourceAttribute::BITS_PER_SAMPLE, fmt::to_string(temp));
+    setBitsPerSample(item, res, aiff);
+}
+
+void TagLibHandler::extractPcm(
+    TagLib::IOStream& roStream,
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res) const
+{
+#if TAGLIB_MAJOR_VERSION >= 2
+    auto wav = TagLib::RIFF::WAV::File(&roStream);
+
+    if (!wav.isValid()) {
+        log_info("TagLibHandler {}: does not appear to be a valid WAV file", item->getLocation().c_str());
+        return;
+    }
+    auto sc = converterManager->i2i();
+    populateGenericTags(item, res, wav, wav.hasID3v2Tag() ? wav.ID3v2Tag()->properties() : wav.properties(), sc);
+
+    setBitsPerSample(item, res, wav);
+#else
+    log_warning("For PCM support in file '{}' TaglibHandler needs to be built with taglib 2 and above", item->getLocation().c_str());
+#endif
+}
+
+template <class Media>
+void TagLibHandler::setBitsPerSample(
+    const std::shared_ptr<CdsItem>& item,
+    const std::shared_ptr<CdsResource>& res,
+    Media& media) const
+{
+    auto audioProps = media.audioProperties();
+    auto bps = audioProps->bitsPerSample();
+    if (bps > 0) {
+        res->addAttribute(ResourceAttribute::BITS_PER_SAMPLE, fmt::to_string(bps));
     }
 }
 
